*  Program...........: XTabGen.prg
*  Author............: Gerald F. Garcia Jr.
*  Project...........: FPCOMMON
*  Created...........: 07/08/1995  06:43:12
*  Copyright.........: (c) 1995
*) Description.......: Cross-tabulates tables.
*)                   : ASSUMES that the table in
*)                   : question is SORTED PROPERLY
*  Calling Samples...: do XTabGen with ;
*)                   :    "", .T., .T., "QUERY", ;
*)                   :    4, 2, .T., .T.
*  Parameter List....: tcOutput
*)                   :    table to output cross-tab
*)                   : tlCursOnly
*)                   :    generate cursor only?
*)                   : tlCloseInp
*)                   :    close input table?
*)                   : tcInput
*)                   :    table to cross-tab
*)                   : tnRowLevel  = 
*)                   :    row "depth"
*)                   :    (ie, grps/subgrps)
*)                   : tnColLevel
*)                   :    column "depth"
*)                   : tlColSubTot
*)                   :    calculate col subtotals?
*)                   : tlColTot
*)                   :    calculate col totals?
*)                   : tcLegend
*)                   :    name of cross-tab legend
*)                   :    to create
*] Major change list.: 09/08/1995-GFG
*)                   :    1. Now handles input
*)                   :       tables with ONLY
*)                   :       three columns.
*)                   :    2. Row and column field(s)
*)                   :       can be of any type
*)                   :       (except Memo, General,
*)                   :       or Picture).
*)                   :    3. No longer chokes when
*)                   :       when tcInput is not
*)                   :       specified

parameters ;
   tcOutput, ;
   tlCursOnly, ;
   tlCloseInp, ;
   tcInput, ;
   tnRowLevel, ;
   tnColLevel, ;
   tlColSubTot, ;
   tlColTot, ;
   tcLegend

#DEFINE dnMaxRowLevels  99
#DEFINE dnMaxColLevels  99
#DEFINE dnMaxFields     255
#DEFINE dnMaxCdxExprLen 240

#DEFINE dcFieldName     1
#DEFINE dcFieldType     2
#DEFINE dnFieldLen      3
#DEFINE dnFieldDec      4

#IF "VISUAL" $ Upper( Version() )
   #DEFINE dlFieldNull    5
   #DEFINE dlFieldNoCp    6
   #DEFINE dcFieldRule    7
   #DEFINE dcFieldText    8
   #DEFINE dcFieldDefault 9
   #DEFINE dcTableRule    10
   #DEFINE dcTableText    11
   
   #DEFINE dnAFields 11
#ELSE
   #UNDEF  dnAFields
   #DEFINE dnAFields 4
#ENDIF
   
private all like j*

private ;
   pcOldTalk, ;
   pcOldExact, ;
   pcOldSafety, ;
   pnOldSelect

*-- Preamble
if Set( 'talk' ) = "ON"
   set talk off
   pcOldTalk = "ON"
else
   pcOldTalk = "OFF"
endif

pcOldExact = Set( 'exact' )
set exact off
pcOldSafety = Set( 'safety' )
set safety off
pnOldSelect = Select( 0 )

*----------------------
*-- VALIDATE PARAMETERS
*----------------------
*-- Validate tcOutput
if Type( "tcOutput" ) = "C"
   tcOutput = Upper( AllTrim( tcOutput ) )
endif

if Empty( tcOutput )
   tcOutput = "XTAB"
endif

*-- Validate tlCursOnly
if Parameters() < 2 or Type( "tlCursOnly" ) # "L"
   tlCursOnly = .T.
endif

*-- Validate tlCloseInp
if Parameters() < 3 or Type( "tlCloseInp" ) # "L"
   tlCloseInp = .T.
endif

*-- Validate tcInput
*{ CHANGE - GFG - September 08, 1995 - 10:47:27
if Type( "tcInput" ) != "C"
   tcInput = ""
endif

tcInput = Upper( AllTrim( tcInput ) )
   
if Empty( tcInput )
   *-- No alias specified,
   *-- assume current work area
   if Empty( Alias() )
      = BailOut( "No table specified" )
      RETURN
   else
      tcInput = Upper( AllTrim( Alias() ) )
   endif
else
   if not Used( tcInput )
      = BailOut( "Table not found" )
      RETURN
   else
      select ( tcInput )
   endif
endif
*}

if tcInput == tcOutput
   = BailOut( ;
      "Input and output tables must be different" ;
      )
   RETURN
endif

*-- Validate tnRowLevel and tnColLevel
if not Type( "tnRowLevel" ) = "N"
   tnRowLevel = 1
else
   if not Between( tnRowLevel, 1, dnMaxRowLevels )
      = BailOut( ;
         "Row levels must be between 1 and " ;
         + LTrim( Str( dnMaxRowLevels ) ) ;
         )
      RETURN
   endif
endif

if not Type( "tnColLevel" ) = "N"
   tnColLevel = 1
else
   if not Between( tnColLevel, 1, dnMaxColLevels )
      = BailOut( ;
         "Column levels must be between 1 and " ;
         + LTrim( Str( dnMaxColLevels ) ) ;
         )
      RETURN
   endif
endif

if not ( ;
   Type( "tnRowLevel" ) = "N" and tnRowLevel >= 1 ;
   and Type( "tnColLevel" ) = "N" ;
   and ( tnRowLevel + tnColLevel + 1 ) = FCount() ;
   )
   = BailOut( "Invalid row/col depths" )
   RETURN
endif

*-- Validate tlColSubTot
private ;
   lcCellType

lcCellType = Type( Field( FCount() ) )

if Type( "tlColSubTot" ) = "L" and tlColSubTot
   if tnColLevel > 1 and lcCellType = "N"
      tlColSubTot = .T.
   else
      *-- Automatically disable column sub-total
      tlColSubTot = .F.
   endif
else
   tlColSubTot = .F.
endif

*-- Validate tlColTot
if Type( "tlColTot" ) = "L" and tlColTot ;
   and lcCellType = "N"
   
   tlColTot = .T.
else
   tlColTot = .F.
endif

*-- Validate tcLegend
if Type( "tcLegend" ) = "C"
   tcLegend = Upper( AllTrim( tcLegend ) )

   if Empty( tcLegend )
      tcLegend = "XTABLGND"
   endif
else
   tcLegend = "XTABLGND"
endif

if tcLegend == tcInput or tcLegend == tcOutput
   = BailOut( ;
      "Input, output and legend tables " ;
      + "must be different" ;
      )
   RETURN
endif

*-- Validate input table
if RecCount() = 0
   = BailOut( "Input table empty" )
   RETURN
else
   jnBadFields = 0

   for jnFldPos = 1 to FCount()
      if Type( Field( jnFldPos ) ) $ "MGP"
         jnBadFields = jnBadFields + 1
      endif
   endfor
   
   if jnBadFields > 0
      = BailOut( ;
         "Input table cannot contain Memo, " ;
         + "General or Picture type fields" ;
         )
      RETURN
   endif
endif

private ;
   pnTotalRecs, ;
   pnRecsDone
   
pnTotalRecs = ( RecCount() * 2 )
pnRecsDone  = 0

*------------------------
*-- OK, LET'S GET STARTED
*------------------------

*-- Copy input table's fields into array
= AFields( laXTabFlds )

*-- Build "AFields()" array for "XTABLGND"
dimension laLgndFlds[ 1, dnAFields ]

laLgndFlds[ 1, dcFieldName ] = "CFIELDNAME"
laLgndFlds[ 1, dcFieldType ] = "C"
laLgndFlds[ 1, dnFieldLen  ] = 10
laLgndFlds[ 1, dnFieldDec  ] = 0

#IF "VISUAL" $ Upper( Version() )
   laLgndFlds[ 1, dlFieldNull ]    = .T.
   laLgndFlds[ 1, dlFieldNoCp ]    = .F.
   laLgndFlds[ 1, dcFieldRule ]    = ""
   laLgndFlds[ 1, dcFieldText ]    = ""
   laLgndFlds[ 1, dcFieldDefault ] = ""
   laLgndFlds[ 1, dcTableRule ]    = ""
   laLgndFlds[ 1, dcTableText ]    = ""
#ENDIF

private ;
   lcIndexExpr, ;
   lcSrchExpr

lcIndexExpr = ""
lcSrchExpr  = ""

for jnLevel = 1 to tnColLevel
   jnALen = ALen( laLgndFlds, 1 ) + 1
   
   dimension laLgndFlds[ jnALen, dnAFields ]

   laLgndFlds[ jnALen, dcFieldType ] = ;
      laXTabFlds[ ;
         tnRowLevel + jnLevel, ;
         dcFieldType ;
         ]
   laLgndFlds[ jnALen, dcFieldName ] = ;
      laLgndFlds[ jnALen, dcFieldType ] ;
      + "LEVEL" + Transform( jnLevel, "@L 99" )
   laLgndFlds[ jnALen, dnFieldLen ] = ;
      laXTabFlds[ ;
         tnRowLevel + jnLevel, ;
         dnFieldLen ;
         ]
   laLgndFlds[ jnALen, dnFieldDec ] = ;
      laXTabFlds[ ;
         tnRowLevel + jnLevel, ;
         dnFieldDec ;
         ]
   #IF "VISUAL" $ Upper( Version() )
      laLgndFlds[ jnALen, dlFieldNull ]    = .T.
      laLgndFlds[ jnALen, dlFieldNoCp ]    = .F.
      laLgndFlds[ jnALen, dcFieldRule ]    = ""
      laLgndFlds[ jnALen, dcFieldText ]    = ""
      laLgndFlds[ jnALen, dcFieldDefault ] = ""
      laLgndFlds[ jnALen, dcTableRule ]    = ""
      laLgndFlds[ jnALen, dcTableText ]    = ""
   #ENDIF

   *-- Build index expression for XTABLGNDWK
   lcIndexExpr = ;
      lcIndexExpr ;
      + XtoCExpr( ;
         laXTabFlds[ ;
            tnRowLevel + jnLevel, ;
            dcFieldName ;
            ], ;
         laXTabFlds[ ;
            tnRowLevel + jnLevel, ;
            dnFieldLen ;
            ], ;
         laXTabFlds[ ;
            tnRowLevel + jnLevel, ;
            dnFieldDec ;
            ], ;
         laLgndFlds[ jnALen, dcFieldType ] ;
            + "LEVEL" ;
            + Transform( jnLevel, "@L 99" ) ;
         ) ;
      + "+"

   *-- Build search expression for Input table
   lcSrchExpr = ;
      lcSrchExpr ;
      + XtoCExpr( ;
         tcInput ;
            + "." ;
            + laXTabFlds[ ;
               tnRowLevel + jnLevel, ;
               dcFieldName ;
               ], ;
         laXTabFlds[ ;
            tnRowLevel + jnLevel, ;
            dnFieldLen ;
            ], ;
         laXTabFlds[ ;
            tnRowLevel + jnLevel, ;
            dnFieldDec ;
            ] ;
         ) ;
      + "+"
endfor      

*-- Strip trailing "+"
lcIndexExpr = ;
   Left( lcIndexExpr, Len( lcIndexExpr ) - 1 )
lcSrchExpr = ;
   Left( lcSrchExpr, Len( lcSrchExpr ) - 1 )

goto top in ( tcInput )   

if Len( Evaluate( lcSrchExpr ) ) > dnMaxCdxExprLen
   = BailOut( ;
      "Too many column levels " ;
      + "(generated index expression " ;
      + "too long)" ;
      )
   RETURN
endif

*-- Create XTABLGND (cross-tab legend)
select 0
create cursor ( tcLegend ) from array laLgndFlds
index on CFIELDNAME tag CFIELDNAME
index on &lcIndexExpr tag SRCHEXPR
index ;
   on Left( CFIELDNAME, 7 ) + &lcIndexExpr ;
   tag SUBTOTSRCH

set order to 0

*-- Create XTABLGNDWK (scratch cursor
*-- used to build XTABLGND). This cursor
*-- is the same as XTABLGND sans CFIELDNAME field.
= ACopy( laLgndFlds, laLgndWork )
= ADel( laLgndWork, 1 )

dimension ;
   laLgndWork[ ;
      ALen( laLgndFlds, 1 ) - 1, ;
      dnAFields ;
      ]
select 0
create cursor XTABLGNDWK from array laLgndWork
index on &lcIndexExpr tag SRCHEXPR

*-- Find unique columns
select ( tcInput )

*-- Insert unique column values into XTABLGNDWK
dimension laColValues[ tnColLevel ]

scan
   if not Seek( ;
      Evaluate( lcSrchExpr ), ;
      "XTABLGNDWK" ;
      )
      *-- Unique, so record it
      for jnLevel = 1 to tnColLevel
         laColValues[ jnLevel ] = ;
            Evaluate( ;
               tcInput + "." ;
               + laXTabFlds[ ;
                  tnRowLevel + jnLevel, ;
                  dcFieldName ;
                  ] ;
               )
      endfor

      insert into XTABLGNDWK from array laColValues
   endif

   pnRecsDone = pnRecsDone + 1
   = ShoPctDone()
endscan

*-- Transfer contents of XTABLGNDWK into XTABLGND
*-- and assign column numbers.
select XTABLGNDWK
goto top

*{ CHANGE - GFG - September 08, 1995 - 10:48:26
dimension laLastGroup[ Max( 1, tnColLevel - 1 ) ]

for jnLevel = 1 to ( Max( 1, tnColLevel - 1 ) )
*}
   laLastGroup[ jnLevel ] = ;
      Evaluate( "XTABLGNDWK." + Field( jnLevel ) )
endfor

private ;
   lnCellLen, ;
   lnCellDec, ;
   lnColumnCnt
   
lnCellLen = ;
   laXTabFlds[ FCount( tcInput ), dnFieldLen ]
lnCellDec = ;
   laXTabFlds[ FCount( tcInput ), dnFieldDec ]
lnColumnCnt = 0

*-- Keep only the "row" fields
dimension laXTabFlds[ tnRowLevel, dnAFields ]

*-- Create array to hold field values for XTABLGND
dimension laColValues[ tnColLevel + 1 ]

scan   
   if tlColSubTot
      *-- Column subtotals desired
      *-- so check for group breaks.
      for jnLevel = ( tnColLevel - 1 ) to 1 step -1
         if Evaluate( ;
            "XTABLGNDWK." + Field( jnLevel ) ;
            ) != laLastGroup[ jnLevel ]

            *-- Break in group
            lnColumnCnt = lnColumnCnt + 1

            store "" to laColValues
            
            laColValues[ 1 ] = ;
               "NSUB" ;
               + Transform( jnLevel, "@L 99" ) ;
               + "C" ;
               + Transform( lnColumnCnt, "@L 999" )

            for jnSubscript = jnLevel to 1 step -1
               laColValues[ jnSubscript + 1 ] = ;
                  laLastGroup[ jnSubscript ]
            endfor
         
            *-- Reset last group var.
            laLastGroup[ jnLevel ] = ;
               Evaluate( ;
                  "XTABLGNDWK." + Field( jnLevel );
                  )
            insert ;
               into ( tcLegend ) ;
               from array laColValues
                  
            = UpdAFields( ;
               @laXTabFlds, ;
               laColValues[ 1 ], ;
               lcCellType, ;
               20, ;
               lnCellDec ;
               )
         endif
      endfor
   endif

   lnColumnCnt = lnColumnCnt + 1

   store "" to laColValues

   laColValues[ 1 ] = ;
      lcCellType ;
      + "COLUMN" ;
      + Transform( lnColumnCnt, "@L 999" )

   for jnLevel = 1 to tnColLevel
      laColValues[ jnLevel + 1 ] = ;
         Evaluate( Field( jnLevel ) )
   endfor
   
   insert into ( tcLegend ) from array laColValues

   = UpdAFields( ;
      @laXTabFlds, ;
      laColValues[ 1 ], ;
      lcCellType, ;
      lnCellLen, ;
      lnCellDec ;
      )
endscan

if tlColSubTot
   for jnLevel = ( tnColLevel - 1 ) to 1 step -1
      lnColumnCnt = lnColumnCnt + 1

      store "" to laColValues
            
      laColValues[ 1 ] = ;
         "NSUB" + Transform( jnLevel, "@L 99" ) ;
         + "C" + Transform( lnColumnCnt, "@L 999" )

      for jnSubscript = jnLevel to 1 step -1
         laColValues[ jnSubscript + 1 ] = ;
            laLastGroup[ jnSubscript ]
      endfor
         
      insert into ( tcLegend ) ;
         from array laColValues
                  
      = UpdAFields( ;
         @laXTabFlds, ;
         laColValues[ 1 ], ;
         lcCellType, ;
         20, ;
         lnCellDec ;
         )
   endfor
endif

if tlColTot
   lnColumnCnt = lnColumnCnt + 1

   jcFieldName = ;
      "NTOTALC" ;
      + Transform( lnColumnCnt, "@L 999" )
      
   insert into ( tcLegend ) ( ;
         CFIELDNAME ;
         ) ;
      values ( ;
         jcFieldName ;
         )
   = UpdAFields( ;
      @laXTabFlds, ;
      jcFieldName, ;
      lcCellType, ;
      20, ;
      lnCellDec ;
      )
endif

if ( lnColumnCnt + tnRowLevel ) > dnMaxFields
   use in XTABLGND
   use in XTABLGNDWK
   = BailOut( "Too many fields generated" )
   RETURN
endif

*-- Kill the XTABLGND scratch cursor
use in XTABLGNDWK

*------------------------------------------------
*-- BUILD CROSS-TABULATED TABLE
*--
*-- WARNING: If a table with the same alias as
*--          tcOutput is open, it WILL BE CLOSED.
*------------------------------------------------
select 0

if Used( "tcOutput" )
   use in ( tcOutput )
endif

if tlCursOnly
   create cursor ( tcOutput ) from array laXTabFlds
else
   *-- WARNING: Safety is off
   create table ( tcOutput ) from array laXTabFlds
endif

select ( tcInput )
goto top

*-- Build row group expression
jcRowGrpExpr = ""

for jnField = 1 to tnRowLevel
   jcRowGrpExpr = ;
      jcRowGrpExpr ;
      + XtoCExpr( ;
         tcInput + "." ;
            + laXTabFlds[ jnField, dcFieldName ], ;
         laXTabFlds[ jnField, dnFieldLen ], ;
         laXTabFlds[ jnField, dnFieldDec ] ;
         ) ;
      + "+"
endfor

*-- Strip trailing "+"
jcRowGrpExpr = ;
   Left( jcRowGrpExpr, Len( jcRowGrpExpr ) - 1 )

*-- Array to hold XTAB field values
*{ CHANGE - GFG - September 08, 1995 - 10:49:17
dimension ;
   laXTabValues[ FCount( tcOutput ) ], ;
   laXTabSubTot[ Max( 1, tnColLevel - 1 ) ]
*}
jnXTabTotal = 0

store 0 to laXTabSubTot

do while not Eof()
   jcLastRowGrp = Evaluate( jcRowGrpExpr )

   *-- Fill array that holds XTAB field values
   *-- with "NULLs" or appropriate data-types.

   *? Change "Visual!" to "Visual" (below)
   *? to enable VFPs .NULL. support.
   #IF "Visual!" $ Version()
      store .NULL. to laXTabValues
   #ELSE
      do case
      case InList( lcCellType, "N", "F" )
         *? Logical converts to "NULL numeric"
         store .F. to laXTabValues
      case lcCellType = "C"
         store "" to laXTabValues
      case lcCellType = "D"
         store {} to laXTabValues
      case lcCellType = "L"
         *? Date converts to "NULL logical"
         store {} to laXTabValues
      endcase
   #ENDIF

   for jnLevel = 1 to ( tnColLevel - 1 )
      laLastGroup[ jnLevel ] = ;
         Evaluate( Field( jnLevel + tnRowLevel ) )
   endfor

   for jnField = 1 to tnRowLevel
      laXTabValues[ jnField ] = ;
         Evaluate( Field( jnField ) )
   endfor   

   do while ;
      Evaluate( jcRowGrpExpr ) = jcLastRowGrp ;
      and not Eof()
      
      select ( tcLegend )

      jcLookup = ;
         Lookup( ;
            CFIELDNAME, ;
            Evaluate( lcSrchExpr ), ;
            CFIELDNAME, ;
            "SrchExpr" ;
            )
      select ( tcInput )
            
      jnSubscript = ;
         Val( SubStr( jcLookup, 8, 3 ) ) ;
         + tnRowLevel
      
      laXTabValues[ jnSubscript ] = ;
         Evaluate( Field( FCount() ) )

      if tlColTot
         *-- Calculate row total
         jnXTabTotal = ;
            jnXTabTotal ;
            + laXTabValues[ jnSubscript ]
      endif
      
      if tlColSubTot

         *-- Calculate row subtotals
         for jnLevel = ;
            ( tnColLevel - 1 ) to 1 step -1

            if Evaluate( ;
               Field( jnLevel + tnRowLevel ) ;
               ) != laLastGroup[ jnLevel ]
               
               *-- Break in group
               jcSubSrch = ;
                  "NSUB" ;
                  + Transform( jnLevel, "@L 99" ) ;
                  + "C"
*{ CHANGE - GFG - September 08, 1995 - 10:49:52
               for jnSubGroup = 1 to jnLevel
                  jcSubSrch = ;
                     jcSubSrch ;
                     + Eval( ;
                        XtoCExpr( ;
                           "laLastGroup" ;
                              + "[ jnSubGroup ]", ;
                           laLgndFlds[ ;
                              jnSubGroup + 1, ;
                              dnFieldLen ;
                              ], ;
                           laLgndFlds[ ;
                              jnSubGroup + 1, ;
                              dnFieldDec ;
                              ] ;
                           ) ;
                        )
               endfor
*}
               select ( tcLegend )
               
               jcLookup = ;
                  Lookup( ;
                     CFIELDNAME, ;
                     jcSubSrch, ;
                     CFIELDNAME, ;
                     "SubTotSrch" ;
                     )
               select ( tcInput )

               laLastGroup[ jnLevel ] = ;
                  Evaluate( ;
                     Field( jnLevel + tnRowLevel );
                     )
               jnSubscript = ;
                  Val( SubStr( jcLookup, 8, 3 ) ) ;
                  + tnRowLevel

               laXTabValues[ jnSubscript ] = ;
                  laXTabSubtot[ jnLevel ]
                  
               laXTabSubtot[ jnLevel ] = 0
            endif

            laXTabSubtot[ jnLevel ] = ;
               laXTabSubtot[ jnLevel ] ;
               + Evaluate( Field( FCount() ) )
         endfor
      endif

      pnRecsDone = pnRecsDone + 1
      = ShoPctDone()
      skip 1
   enddo

   *-- If specified, put subtotals for 
   *-- "trailing" records
   if tlColSubtot
      for jnLevel = ( tnColLevel - 1 ) to 1 step -1
         jcSubSrch = ;
            "NSUB" ;
            + Transform( jnLevel, "@L 99" ) ;
            + "C"
*{ CHANGE - GFG - September 08, 1995 - 10:50:15
         for jnSubGroup = 1 to jnLevel
            jcSubSrch = ;
               jcSubSrch ;
               + Eval( ;
                  XtoCExpr( ;
                     "laLastGroup[ jnSubGroup ]", ;
                     laLgndFlds[ ;
                        jnSubGroup + 1, ;
                        dnFieldLen ;
                        ], ;
                     laLgndFlds[ ;
                        jnSubGroup + 1, ;
                        dnFieldDec ;
                        ] ;
                     ) ;
                  )
         endfor
*}
         select ( tcLegend )
               
         jcLookup = ;
            Lookup( ;
               CFIELDNAME, ;
               jcSubSrch, ;
               CFIELDNAME, ;
               "SubTotSrch" ;
               )
         select ( tcInput )

         jnSubscript = ;
            Val( SubStr( jcLookup, 8, 3 ) ) ;
            + tnRowLevel

         laXTabValues[ jnSubscript ] = ;
            laXTabSubtot[ jnLevel ]
                  
         laXTabSubtot[ jnLevel ] = 0
      endfor

      store "" to laLastGroup
   endif

   *-- Row total specified, insert it
   if tlColTot
      laXTabValues[ FCount( tcOutput ) ] = ;
         jnXTabTotal

      jnXTabTotal = 0
   endif
   
   insert into ( tcOutput ) from array laXTabValues
enddo

wait clear

*-- Clean up
= BailOut()

if tlCloseInp
   select ( tcOutput )
   use in ( tcInput )
endif

RETURN


************************************************
FUNCTION BailOut    
************************************************

*) Function..........: BailOut    
*  Author............: Gerald F. Garcia Jr.
*  Project...........: FPCOMMON
*  Created...........: 07/08/1995  07:44:40
*  Copyright.........: (c) 1995
*) Description.......: 
*  Calling Samples...: = BailOut( ;
*)                   :    "Unexpected error" ;
*)                   :    )
*  Parameter List....: tcMsg
*)                   :    error message to display
*] Major change list.: 
 
parameters ;
   tcMsg
   
if not Empty( tcMsg )
   ?? Chr( 7 )
   wait window nowait tcMsg
endif

select ( pnOldSelect )

if pcOldExact = "ON"
   set exact on
endif

if pcOldSafety = "ON"
   set safety on
endif

if pcOldTalk = "ON"
   set talk on
endif

RETURN 


************************************************
FUNCTION UpdAFields
************************************************

*) Function..........: UpdAFields
*  Author............: Gerald F. Garcia Jr.
*  Project...........: FPCOMMON
*  Created...........: 07/08/1995  08:32:35
*  Copyright.........: (c) 1995
*) Description.......: Adds row of field info
*)                   : to an "AFIELDS()" array.
*  Calling Samples...: = UpdAFields( ;
*)                   :    @laFields, ;
*)                   :    "DDATE", "D", 8, 0 ;
*)                   :    )
*  Parameter List....: taFields
*)                   :    array passed by REFERENCE
*)                   : tcFieldName
*)                   :    name of field to
*)                   :    add to array
*)                   : tcFieldType
*)                   : tnFieldLen
*)                   : tnFieldDec
*] Major change list.: 
 
parameters ;
   taFields, ;
   tcFieldName, ;
   tcFieldType, ;
   tnFieldLen, ;
   tnFieldDec
   
external array taFields

private all like j*

if Empty( tcFieldType )
   tcFieldType = "C"
endif

if Empty( tnFieldLen )
   do case
   case tcFieldType = "D"
      tnFieldLen = 8
   case tcFieldType = "L"
      tnFieldLen = 1
   otherwise
      tnFieldLen = 10
   endcase
endif

if Empty( tnFieldDec )
   tnFieldDec = 0
endif
 
jnALen = ALen( taFields, 1 ) + 1

dimension taFields[ jnALen, dnAFields ]

taFields[ jnALen, dcFieldName ] = tcFieldName
taFields[ jnALen, dcFieldType ] = tcFieldType
taFields[ jnALen, dnFieldLen ]  = tnFieldLen
taFields[ jnALen, dnFieldDec ]  = tnFieldDec

#IF "VISUAL" $ Upper( Version() )
   taFields[ jnALen, dlFieldNull ]    = .T.
   taFields[ jnALen, dlFieldNoCp ]    = .F.
   taFields[ jnALen, dcFieldRule ]    = ""
   taFields[ jnALen, dcFieldText ]    = ""
   taFields[ jnALen, dcFieldDefault ] = ""
   taFields[ jnALen, dcTableRule ]    = ""
   taFields[ jnALen, dcTableText ]    = ""
#ENDIF

RETURN


************************************************
FUNCTION XtoCExpr  
************************************************

*) Function..........: XtoCExpr  
*  Author............: Gerald F. Garcia Jr.
*  Project...........: FPCOMMON
*  Created...........: 07/08/1995  08:35:03
*  Copyright.........: (c) 1995
*) Description.......: Returns expression to
*)                   : convert field to char-type
*  Calling Samples...: jcExpr = jcExpr ;
*)                   :    + XtoCExpr( ;
*)                   :       "DDATE", ;
*)                   :       8, ;
*)                   :       0, ;
*)                   :       "PROCDATE" ;
*)                   :       )
*  Parameter List....: tcFieldName
*)                   : tnFieldLen
*)                   : tnFieldDec
*)                   : tcFieldSubst
*)                   :    substitute field name
*] Major change list.: 

parameters ;
   tcFieldName, ;
   tnFieldLen, ;
   tnFieldDec, ;
   tcFieldSubst
   
tcFieldName = Upper( tcFieldName )

if Empty( tcFieldSubst )
   tcFieldSubst = tcFieldName
else
   tcFieldSubst = Upper( tcFieldSubst )
endif

if Empty( tnFieldLen )
   tnFieldLen = 0
endif

if Empty( tnFieldDec )
   tnFieldDec = 0
endif
*{ CHANGE - GFG - September 08, 1995 - 10:50:37 
private ;
   jcRet, ;
   jcFieldType, ;
   jcFieldLen, ;
   jcFieldDec
*}
jcRet       = ""
jcFieldType = Type( tcFieldName )
jcFieldLen  = AllTrim( Str( tnFieldLen ) )
jcFieldDec  = AllTrim( Str( tnFieldDec ) )
 
do case
case jcFieldType = "C"
   jcRet = tcFieldSubst
case jcFieldType = "D"
   jcRet = "DTOS(" + tcFieldSubst + ")"
case InList( jcFieldType, "F", "N" )
   jcRet = ;
      "STR(" + tcFieldSubst + "," ;
      + jcFieldLen + "," + jcFieldDec + ")"
case jcFieldType = "L"
   jcRet = "IIF(" + tcFieldSubst + ",'T','F')"
otherwise
   jcRet = ""
endcase

RETURN ( jcRet )


************************************************
FUNCTION ShoPctDone
************************************************

*) Function..........: ShoPctDone
*  Author............: Gerald F. Garcia Jr.
*  Project...........: FPCOMMON
*  Created...........: 07/08/1995  15:20:38
*  Copyright.........: (c) 1995
*) Description.......: Displays progress indicator
*  Calling Samples...: = ShoPctDone()
*  Parameter List....: 
*] Major change list.: 
 
private all like j*
 
jcPctDone = ;
   Str( ;
      Round( ;
         ( pnRecsDone / pnTotalRecs ) * 100, ;
         0 ;
         ), ;
      3, ;
      0 ;
      )
wait window nowait ;
   "Generating cross-tab: " ;
   + jcPctDone ;
   + "% completed"

RETURN